#include "NF.h"

extern bool fap, pfp;
extern flow Fmax;
int FIFO = 0;
//
//	NETWORK FLOW COMPUTE
//
//  FLOW AUGMENTING PATH (FAP) METHOD  
//
//  Find path from sv to tv in residual network, using non saturated forward arcs and positive-flow backward arcs  
//
 void
    NF_distances( digraph &NG, arcs &NA, vertex sv, vertex tv, lengths &ND ) 
{
    vertex vn = NG.size();
	arc an = NA.size() / 2;
//  Vertices list
	vector<vertex> Nv(vn);
	vector<bool> Nl(vn, false);
//  List index
	index BFSn = 0, BFSi = 0;
	Nv[BFSn++] = tv;
	Nl[tv] = false;
	ND[tv] = 0;
	do {
	//  New vertex v to display 
		vertex v = Nv[BFSi++];
		for( index i=0; i<NG[v].size(); i++ ) 
			if (!Nl[NG[v][i]]) {
				arc a = NA[vip(v, i)];
				if (a>=an) {
					Nv[BFSn++] = NG[v][i];
					Nl[NG[v][i]] = true;
					ND[NG[v][i]] = ND[v] + 1;
				}
			}
	} while( BFSi < BFSn );
	ND[sv] = vn;
	return;
}

 bool
	 NF_Path(digraph& NG, arcs& NA, capacities& NC, flows& NF, vertex sv, vertex tv, vector<vertex>& Nuv, vector<arc>& Nue)
 {
	 vertex vn = NG.size();
	 arc an = NA.size() / 2;
	 //  Vertices list
	 vector<vertex> Nv(vn);
	 //  Upper vertices vector to vn  and upper arcs
	 Nuv = vector<vertex>(vn, vn);
	 Nue = vector<arc>(vn);
	 //  List index
	 index BFSn = 0, BFSi = 0;
	 Nuv[sv] = sv;
	 Nv[BFSn++] = sv;
	 if (NG[sv].size())
		 do {
			 //  New vertex v to display 
			 vertex v = Nv[BFSi++];
			 for (index i = 0; i < NG[v].size(); i++)
				 if (Nuv[NG[v][i]] == vn) {
					 arc a = NA[vip(v, i)];
					 if (a < an && NF[a] < NC[a]) {
						 //  Non saturated forward arc
						 //  Upper vertex and upper arc and list adjacent N[v][i]
						 Nuv[NG[v][i]] = v;
						 Nue[NG[v][i]] = a;
						 Nv[BFSn++] = NG[v][i];
						 if (NG[v][i] == tv)
							 //  Path from sv to tv found!
							 return true;
					 }
					 if (a >= an && NF[a - an] > 0) {
						 //  Positive flow backward arc
						 //  Upper vertex and upper arc and list adjacent N[v][i]
						 Nuv[NG[v][i]] = v;
						 Nue[NG[v][i]] = a;
						 Nv[BFSn++] = NG[v][i];
						 if (NG[v][i] == tv)
							 //  Path from sv to tv found!
							 return true;
					 }
				 }
		 } while (BFSi < BFSn);
		 Nuv = vector<vertex>(0);
		 return false;
 }

 vector<vertex> Nuv; vector<arc> Nue;
//
//	Augment flow through the eventualy found path in the residual network:
//
 bool
	 NF_FlowAugmentingPath( digraph &NG, arcs &NA, capacities &NC, flows &NF, vertex sv, vertex tv )
{
	arc an = NA.size() / 2;
	if (!NF_Path(NG, NA, NC, NF, sv, tv, Nuv, Nue))
		return false;
//  Find admissible incremental flow df through the path: minimum of factible forward increments / backward decrements
	vertex v = tv; arc a;
	int df = 1000;
	while (v != sv) {
		a = Nue[v];
		if ( a<an && df > NC[a] - NF[a])
			df = NC[a] - NF[a];
		else if( a>an && df > NF[a-an] )
			df = NF[a-an];
		v = Nuv[v];
	}
//  Apply admissible incremental flow df to the flow augmenting path:
//  increment df for forward arcs and decrement df for backward arcs
//  Write path to cout
	v = tv;
	cout << tv;
	while (v != sv) {
		a = Nue[v];
		if (a < an)
			NF[a] += df;
		else
			NF[a-an] -= df;
		v = Nuv[v];	
		cout << "-(" << a << ")" << v;
	}
	Fmax += df;
	cout << "[" << df << "]" << endl;
	cout << "[" << Fmax << "]" << endl;
	return true;
}
//
//	Flow Augmenting Path Method for Network Flow Problems
//
 flow
	 NF_FAP_Method( digraph &NG, points &Np, arcs &NA, capacities &NC, flows &NF, vertex sv, vertex tv, ofstream &ftex, ofstream &fout )
 {
	 Fmax = 0;
	 do {
		 NF_tex(NG, Np, NA, NC, NF, sv, tv, ftex);
		 NF_write(NG, Np, NA, NC, NF, sv, tv, fout);
	 } while (NF_FlowAugmentingPath(NG, NA, NC, NF, sv, tv));
	 cout << endl;
	 return Fmax;
 }
 //
 //  PREFLOW PUSH (PFP) METHOD OF 
 //
 //  Preflow initialization: from sv flow to capacity and compute excesses
 //
 void
	 NF_PreflowInit(digraph &NG, arcs &NA, capacities &NC, flows &NF, vertex sv, vertex tv, lengths &ND, excesses &NE)
 {
	 vertex vn = NG.size();
	 arc an = NA.size() / 2;
	 //  Push all flow at sv
	 for (index si = 0; si < NG[sv].size(); si++) {
		 arc a = NA[vip(sv, si)];
		 if (a < an)
			 NF[a] = NC[a];
	 }
	 ND = lengths(vn);
	 NF_distances(NG, NA, sv, tv, ND);
	 //  Excesses
	 NE = excesses(vn, 0);
	 for (vertex v = 0; v<vn; v++)
		 for (index i = 0; i < NG[v].size(); i++) {
			 arc a = NA[vip(v, i)];
			 if (a < an)
				 NE[v] -= NF[a];
			 else
				 NE[v] += NF[a - an];
		 }
 }

 //
 //	 Preflow push from an excess vertex through an admissible vertex
 //
 bool
	 NF_PreflowPush(digraph &NG, arcs &NA, capacities &NC, flows &NF, vertex sv, vertex tv, lengths &ND, excesses &NE)
 {
	 vertex vn = NG.size();
	 arc an = NA.size() / 2;
	 //  Find first excess vertex xv
	 vertex xv;
	 for (xv = 0; xv < vn; xv++)
		 if (NE[xv]>0 && xv != tv)
			 break;
	 cout << "Excess vertex " << xv << endl;
	 if (xv == vn)
		 return false;
	 //  PUSH
	 //  Search one admissible residual arc from/to xv to push flow through it
	 //  Admissible: from/to a residual adjacent 
	 index xi;
	 for (xi = 0; xi < NG[xv].size(); xi++) {
		 arc a = NA[vip(xv, xi)];
		 vertex axv = NG[xv][xi];
		 int rfl;
		 if (a<an && (rfl = NC[a] - NF[a])>0 && ND[xv] == ND[axv] + 1) {
			 //  Forward non-saturated arc: push feasible excess 
			 flow delta = min(NE[xv], rfl);
			 NE[xv] -= delta;
			 NE[axv] += delta;
			 NF[a] += delta;
			 cout << "Push " << delta << " through forward arc " << a << "(" << xv << "," << axv << ")" << endl;
			 break;
		 }
		 if (a >= an && (rfl = NF[a - an])>0 && ND[xv] == ND[axv] + 1) {
			 //  Backward positive-flow arc: push feasible excess
			 flow delta = min(NE[xv], rfl);
			 NE[xv] -= delta;
			 NE[axv] += delta;
			 NF[a - an] -= delta;
			 cout << "Push " << delta << " through backward arc " << a << "(" << xv << "," << axv << ")" << endl;
			 break;
		 }
	 }
	 if (xi == NG[xv].size()) {
		 //  RELABEL
		 //  Relabel distance of xv: minimum distance from/to (residual) adjacent vertices + 1
		 ND[xv] = 1 << 10;
		 for (xi = 0; xi < NG[xv].size(); xi++) {
			 arc a = NA[vip(xv, xi)];
			 vertex axv = NG[xv][xi];
			 if (a < an && NF[a]<NC[a] && ND[xv]>ND[axv] + 1)
				 //  Residual forward non-satutared arc
				 ND[xv] = ND[axv] + 1;
			 if (a >= an && NF[a - an]>0 && ND[xv]>ND[axv] + 1)
				 //  Residual backward positive-flow arc
				 ND[xv] = ND[axv] + 1;
		 }
		 cout << "No arcs admissibles so we relabel the node " << xv << endl;
	 }
	 Fmax = NE[tv];
	 return true;
 }


 //
 //	 FIFO Preflow push from an excess vertex through an admissible vertex
 //
 bool
	 NF_FIFOPreflowPush(digraph& NG, points &Np, arcs& NA, capacities& NC, flows& NF, vertex sv, vertex tv, lengths& ND, excesses& NE, ofstream& ftex, ofstream& fout)
 {
	 vertex vn = NG.size();
	 arc an = NA.size() / 2;
	 vertices NV;
	 vector<bool>Nl(vn,false);
	 //  Find all the vertices xv with excess
	 vertex xv;
	 for (xv = 0; xv < vn; xv++)
		 if (NE[xv] > 0 && xv != tv) {
			 NV.push_back(xv);
			 Nl[xv] = true;
		 }
	 if (NV.size() == 0)
		 return false;
	 cout << "Excess vertices are: (";;
	 for ( xv = 0; xv < NV.size(); xv++)
		 cout << NV[xv] << ",";
	 cout << ")" << endl;
	 //  PUSH
	 //  Search one admissible residual arc from/to xv to push flow through it
	 //  Admissible: from/to a residual adjacent 
	 index xi;
	 do {
		 xv = NV[0];
		 for (xi = 0; xi < NG[xv].size(); xi++) {
			 arc a = NA[vip(xv, xi)];
			 vertex axv = NG[xv][xi];
			 int rfl;
			 if (a < an && (rfl = NC[a] - NF[a])>0 && ND[xv] == ND[axv] + 1) {
				 //  Forward non-saturated arc: push feasible excess 
				 flow delta = min(NE[xv], rfl);
				 NE[xv] -= delta;
				 NE[axv] += delta;
				 NF[a] += delta;
				 cout << "Push " << delta << " through forward arc " << a << "(" << xv << "," << axv << ")" << endl;
				 break;
			 }
			 if (a >= an && (rfl = NF[a - an]) > 0 && ND[xv] == ND[axv] + 1) {
				 //  Backward positive-flow arc: push feasible excess
				 flow delta = min(NE[xv], rfl);
				 NE[xv] -= delta;
				 NE[axv] += delta;
				 NF[a - an] -= delta;
				 cout << "Push " << delta << " through backward arc " << a << "(" << xv << "," << axv << ")" << endl;
				 break;
			 }
		 }
		 if (xi == NG[xv].size()) {
			 //  RELABEL
			 //  Relabel distance of xv: minimum distance from/to (residual) adjacent vertices + 1
			 ND[xv] = 1 << 10;
			 for (xi = 0; xi < NG[xv].size(); xi++) {
				 arc a = NA[vip(xv, xi)];
				 vertex axv = NG[xv][xi];
				 if (a < an && NF[a]<NC[a] && ND[xv]>ND[axv] + 1)
					 //  Residual forward non-satutared arc
					 ND[xv] = ND[axv] + 1;
				 if (a >= an && NF[a - an] > 0 && ND[xv] > ND[axv] + 1)
					 //  Residual backward positive-flow arc
					 ND[xv] = ND[axv] + 1;
			 }
			 cout << "No arcs admissibles so we relabel the node " << xv << endl;
			 NV.erase(NV.begin());
			 NV.push_back(xv);
		 }
		 else for (xv = 0; xv < vn; xv++)
			 if (NE[xv] > 0 && xv != tv && Nl[xv] == false) {
				 NV.push_back(xv);
				 Nl[xv] = true;
			 }
		 xv = NV[0];
		 if (NE[xv] == 0) {
			 NV.erase(NV.begin());
			 Nl[xv] = false;
		 }
		 // CORRECCI xv = NV[0];
		 NF_tex(NG, Np, NA, NC, NF, sv, tv, ftex);
		 NF_write(NG, Np, NA, NC, NF, sv, tv, fout);
	 } while (NV.size());
	 Fmax = NE[tv];
	 return true;
 }
//
//	 Preflow Push Method for Network Flow Problems
//

 flow
	 NF_PFP_Method(digraph &NG, points &Np, arcs &NA, capacities &NC, flows &NF, vertex sv, vertex tv, ofstream &ftex, ofstream &fout)
 {
	 lengths ND; excesses NE;
	 NF_PreflowInit(NG, NA, NC, NF, sv, tv, ND, NE);
	 if( FIFO )
		NF_FIFOPreflowPush(NG, Np, NA, NC, NF, sv, tv, ND, NE, ftex, fout);
	 else do {
		 NF_tex(NG, Np, NA, NC, NF, sv, tv, ftex);
		 NF_write(NG, Np, NA, NC, NF, sv, tv, fout);
	 } while (NF_PreflowPush(NG, NA, NC, NF, sv, tv, ND, NE));
	 return Fmax;
}
 
 
 void
	 NF_Init(digraph &NG, points &Np, arcs &NA, capacities &NC, flows &NF, vertex &sv, vertex &tv, lengths& ND, excesses &NE,
		 ofstream &ftex, ofstream &fout)

 {
	network_Init(NG, Np, NA, NC, NF, sv, tv, ftex, fout);
	if( pfp )
		NF_PreflowInit(NG, NA, NC, NF, sv, tv, ND, NE);
}
 //
 //  Network flow computing iteration
 //
 void
	 NF_Compute(digraph &NG, points &Np, arcs &NA, capacities &NC, flows &NF, vertex &sv, vertex &tv, lengths& ND, excesses &NE,
		 ofstream &ftex, ofstream &fout )
 {
	 if (fap) {
		 NF_tex(NG, Np, NA, NC, NF, sv, tv, ftex);
		 NF_write(NG, Np, NA, NC, NF, sv, tv, fout);
		 fap = NF_FlowAugmentingPath(NG, NA, NC, NF, sv, tv);
	 }
	 else if  (pfp) {
		 NF_tex(NG, Np, NA, NC, NF, sv, tv, ftex);
		 NF_write(NG, Np, NA, NC, NF, sv, tv, fout);
		 pfp = NF_PFP_Method(NG, Np, NA, NC, NF, sv, tv, ftex, fout);
	 }
 }




 //
 //	 Bipartite Preflow push from an excess vertex through an admissible vertex
 //
 bool
	 BPNF_PreflowPush(digraph& NG, arcs& NA, capacities& NC, flows& NF, vertex sv, vertex tv, lengths& ND, excesses& NE)
 {
	 vertex vn = NG.size();
	 arc an = NA.size() / 2;
	 arc a, b;
	 int rfl, rflo;
	 //  Find first excess vertex xv
	 vertex xv, axv, bxv;
	 for (xv = 0; xv < vn; xv++)
		 if (NE[xv] > 0 && xv != tv)
			 break;
	 // aqui es para
	 // cout << "Excess vertex " << xv << endl;
	 if (xv == vn)
		 return false;
	 //  PUSH
	 //  Search one admissible residual arc from/to xv to push flow through it
	 //  Admissible: from/to a residual adjacent 
	 index xi;
	 bool flag = false;
	 for (xi = 0; xi < NG[xv].size(); xi++) {
		 index zi = 0;
		 a = NA[vip(xv, xi)];
		 axv = NG[xv][xi];
		 if (a < an && (rfl = NC[a] - NF[a])>0 && ND[xv] == ND[axv] + 1) {
			 //  Forward non-saturated arc: push feasible excess
			 for (zi = 0; zi < NG[axv].size(); zi++) {
				 b = NA[vip(axv, zi)];
				 bxv = NG[axv][zi];
				 if (b < an && (rflo = NC[b] - NF[b])>0 && ND[axv] == ND[bxv] + 1) {
					 //  Forward non-saturated arc: push feasible excess 
					 flow delta = min(NE[xv], rfl);
					 delta = min(delta, rflo);
					 NE[xv] -= delta;
					 NE[bxv] += delta;
					 NF[a] += delta;
					 NF[b] += delta;
					 cout << "Push " << delta << " through forward arc " << a << "(" << xv << "," << axv << ")" << " and through forward arc " << b << "(" << axv << "," << bxv << ")" << endl;
					 flag = true;
					 break;
				 }
				 if (b >= an && (rflo = NF[b - an]) > 0 && ND[axv] == ND[bxv] + 1) {
					 //  Backward positive-flow arc: push feasible excess
					 flow delta = min(NE[xv], rfl);
					 delta = min(delta, rflo);
					 NE[xv] -= delta;
					 NE[bxv] += delta;
					 NF[a] += delta;
					 NF[b - an] -= delta;
					 cout << "Push " << delta << " through forward arc " << a << "(" << xv << "," << axv << ")" << " and through backward arc " << b << "(" << axv << "," << bxv << ")" << endl;
					 flag = true;
					 break;
				 }
			 }
		 }
		 if (a >= an && (rfl = NF[a - an]) > 0 && ND[xv] == ND[axv] + 1) {
			 //  Backward positive-flow arc: push feasible excess
			 for (zi = 0; zi < NG[axv].size(); zi++) {
				 b = NA[vip(axv, zi)];
				 bxv = NG[axv][zi];
				 if (b < an && (rflo = NC[b] - NF[b])>0 && ND[axv] == ND[bxv] + 1) {
					 //  Forward non-saturated arc: push feasible excess 
					 flow delta = min(NE[xv], rfl);
					 delta = min(delta, rflo);
					 NE[xv] -= delta;
					 NE[bxv] += delta;
					 NF[a - an] -= delta;
					 NF[b] += delta;
					 cout << "Push " << delta << " through backward arc " << a << "(" << xv << "," << axv << ")" << " and through forward arc " << b << "(" << axv << "," << bxv << ")" << endl;
					 flag = true;
					 break;
				 }
				 if (b >= an && (rflo = NF[b - an]) > 0 && ND[axv] == ND[bxv] + 1) {
					 //  Backward positive-flow arc: push feasible excess
					 flow delta = min(NE[xv], rfl);
					 delta = min(delta, rflo);
					 NE[xv] -= delta;
					 NE[bxv] += delta;
					 NF[a - an] -= delta;
					 NF[b - an] -= delta;
					 cout << "Push " << delta << " through backward arc " << a << "(" << xv << "," << axv << ")" << " and through backward arc " << b << "(" << axv << "," << bxv << ")" << endl;
					 flag = true;
					 break;
				 }
			 }
		 }
		 if (zi == NG[axv].size()) {
			 //  RELABEL
			 //  Relabel distance of xv: minimum distance from/to (residual) adjacent vertices + 1
			 ND[axv] = 1 << 10;
			 for (zi = 0; zi < NG[axv].size(); zi++) {
				 b = NA[vip(axv, zi)];
				 bxv = NG[axv][zi];
				 if (b < an && NF[b]<NC[b] && ND[axv]>ND[bxv] + 1) {
					 //  Residual forward non-satutared arc
					 ND[axv] = ND[bxv] + 1;
					 cout << "No arcs admissibles so we relabel the node " << axv << endl;
				 }
				 if (b >= an && NF[b - an] > 0 && ND[axv] > ND[bxv] + 1) {
					 //  Residual backward positive-flow arc
					 ND[axv] = ND[bxv] + 1;
					 cout << "No arcs admissibles so we relabel the node " << axv << endl;
				 }
			 }
		 }
		 if (flag)
			 break;
	 }
	 if (xi == NG[xv].size()) {
		 //  RELABEL
		 //  Relabel distance of xv: minimum distance from/to (residual) adjacent vertices + 1
		 ND[xv] = 1 << 10;
		 for (xi = 0; xi < NG[xv].size(); xi++) {
			 a = NA[vip(xv, xi)];
			 axv = NG[xv][xi];
			 if (a < an && NF[a]<NC[a] && ND[xv]>ND[axv] + 1) {
				 //  Residual forward non-satutared arc
				 ND[xv] = ND[axv] + 1;
				 cout << "No arcs admissibles so we relabel the node " << xv << endl;
			 }
			 if (a >= an && NF[a - an] > 0 && ND[xv] > ND[axv] + 1) {
				 //  Residual backward positive-flow arc
				 ND[xv] = ND[axv] + 1;
				 cout << "No arcs admissibles so we relabel the node " << xv << endl;
			 }
		 }
	 }
	 Fmax = NE[tv];
	 return true;
 }






 void
	 BPNF_Init(digraph& NG, points& Np, arcs& NA, capacities& NC, flows& NF, vertex& sv, vertex& tv, lengths& ND, excesses& NE,
		 ofstream& ftex, ofstream& fout)
 {
	 bpnetwork_Init(NG, Np, NA, NC, NF, sv, tv, ftex, fout);
	 if (pfp)
		 NF_PreflowInit(NG, NA, NC, NF, sv, tv, ND, NE);
 }

 flow
	 BPNF_PFP_Method(digraph& NG, points& Np, arcs& NA, capacities& NC, flows& NF, vertex sv, vertex tv, lengths &ND, excesses &NE, ofstream& ftex, ofstream& fout)
 {
	 /*NF_PreflowInit(NG, NA, NC, NF, sv, tv, ND, NE);
	 do {
		 NF_tex(NG, Np, NA, NC, NF, sv, tv, ftex);
		 NF_write(NG, Np, NA, NC, NF, sv, tv, fout);
	 } while (BPNF_PreflowPush(NG, NA, NC, NF, sv, tv, ND, NE));*/
	 NF_tex(NG, Np, NA, NC, NF, sv, tv, ftex);
	 NF_write(NG, Np, NA, NC, NF, sv, tv, fout);
	 BPNF_PreflowPush(NG, NA, NC, NF, sv, tv, ND, NE);
	 return Fmax;
 }
 //
 //  Network flow computing iteration
 //
 void
	 BPNF_Compute(digraph& NG, points& Np, arcs& NA, capacities& NC, flows& NF, vertex& sv, vertex& tv, lengths& ND, excesses& NE,
		 ofstream& ftex, ofstream& fout)
 {
	 if (fap) {
		 NF_tex(NG, Np, NA, NC, NF, sv, tv, ftex);
		 NF_write(NG, Np, NA, NC, NF, sv, tv, fout);
		 fap = NF_FlowAugmentingPath(NG, NA, NC, NF, sv, tv);
	 }
	 else if (pfp) {
		 NF_tex(NG, Np, NA, NC, NF, sv, tv, ftex);
		 NF_write(NG, Np, NA, NC, NF, sv, tv, fout);
		 pfp = BPNF_PFP_Method(NG, Np, NA, NC, NF, sv, tv, ND, NE, ftex, fout);
	 }
 }


