#include "NF.h"
extern int DIGRAPH;
extern string NFstr, BPNFstr;

// 
//  Read graph from file fname 
//  to adjacency lists G, points Gp, edges GE
//
 void
	graph_read( graph &G, points& Gp, string fname)
{
	ifstream fin;
	fin.open(fname.c_str());
	if (fin.fail()) {
		cerr << "unable to open file " << fname.c_str() << " for reading" << endl;
		exit(1);
	}
	vertex vn;
	edge en;
	fin >> vn >> en;
	G = graph(vn);
	Gp = points(vn);
	for (index vi = 0; vi<vn; vi++) {
		vertex v;
		double px, py;
		fin >> v >> px >> py;
		Gp[v].first = px;
		Gp[v].second = py;
	}
	for( edge ei = 0; ei<en; ei++) {
		vertex u, v;
		capacity c;
		edge e;
		fin >> e >> u >> v;
		G[u].push_back(v);
		if (u != v)
			G[v].push_back(u);
	}
}
 // 
 //  Read digraph from file fname 
 //  to adjacency lists G, points Gp, arcs GA
 //
 void
	digraph_read(graph &G, points& Gp, arcs &GA, string fname)
{
	ifstream fin;
	fin.open(fname.c_str());
	if (fin.fail()) {
		cerr << "unable to open file " << fname.c_str() << " for reading" << endl;
		exit(1);
	}
	vertex vn;
	arc an;
	fin >> vn >> an;
	G = graph(vn);
	Gp = points(vn);
	GA.clear();
	for (index vi = 0; vi<vn; vi++) {
		vertex v;
		double px, py;
		fin >> v >> px >> py;
		Gp[v].first = px;
		Gp[v].second = py;
	}
	for (arc ai = 0; ai<an; ai++) {
		vertex u, v;
		arc a;
		fin >> a >> u >> v;
		GA[vip(u, G[u].size())] = ai;
		if (u != v)
			GA[vip(v, G[v].size())] = ai + an*DIGRAPH;
		G[u].push_back(v);
		if (u != v)
			G[v].push_back(u);
	}
}

extern string DGstr;
 void
	 digraph_Init(digraph &G, points &Gp, arcs &GA, ofstream &ftex, ofstream &fout)
{
	 string DGdat = DGstr;
	 DGdat += ".dat";
	 digraph_read(G, Gp, GA, DGdat);
	 string DGout = DGstr;
	 DGout += ".out";
	 string DGtex = DGstr;
	 DGtex += ".tex";
	 fout.open(DGout);
	 ftex.open(DGtex);
	 digraph_write(G, Gp, GA, fout);
	 digraph_tex(G, Gp, GA, ftex);
}



// 
//  Read Network Flow Problem from file fname 
//  to adjacency lists NG, points Np, arcs NA, capacities NC, flows NF
//
 void
	NF_read( digraph &NG, points& Np, arcs &NA, capacities &NC, flows &NF, vertex &sv, vertex &tv, string fname )
{
	ifstream fin;
	fin.open(fname.c_str());
	if (fin.fail()) {
		cerr << "unable to open file " << fname.c_str() << " for reading" << endl;
		exit(1);
	}
	vertex vn;
	arc an;
	fin >> vn >> an;
	NG = graph(vn);
	Np = points(vn);
	NC = capacities(an);
	NF = flows(an,0);
	NA.clear();
	for (index vi = 0; vi<vn; vi++) {
		vertex v;
		double px, py;
		fin >> v >> px >> py;
		Np[v].first = px;
		Np[v].second = py;
	}
	for (arc ai = 0; ai<an; ai++) {
		vertex u, v;
		capacity c;
		arc a;
		fin >> a >> u >> v >> c;
		NA[vip(u, NG[u].size())] = ai;
		if (u != v)
			NA[vip(v, NG[v].size())] = ai+an;
		NG[u].push_back(v);
		if (u != v)
			NG[v].push_back(u);
		NC[ai] = c;
	}
	fin >> sv >> tv;
}
extern int Fmax;
// 
//  Write Network Flow to fout
//
void
	NF_write( digraph& NG, points &Np, arcs &NA, capacities &NC, flows &NF, vertex sv, vertex tv, ofstream& fout )
{
	//  Vertices number 
	vertex vn = NG.size();
	//  Count vips number from adjacencies lists
	arc an = 0;
	for (vertex v = 0; v < vn; v++)
		an = an + NG[v].size();
	an /= 2;
	//  Write vips         
	fout << "Network with " << vn << " vertices and " << an << " arcs " << endl;
	fout << "Adjacencies lists" << endl;
	for (vertex v = 0; v<vn; v++) {
		flow vf = 0;
		fout << v << "\t:\t";
		for (index i = 0; i<NG[v].size(); i++) {
			arc a = NA[vip(v, i)];
			fout << "\t" << NG[v][i] << " (" << a << ")";
			if (a < an) {
				fout << "[" << NC[a] << "/" << NF[a] << "]";
				vf += NF[a];
			}
			else
				vf -= NF[a - an];
		}
		fout << "     " << "{" << vf << "}" << endl;
	}
	fout << "Arcs" << endl;
	for (vertex v = 0; v < vn; v++)
		for (index i = 0; i < NG[v].size(); i++) {
			arc a = NA[vip(v, i)];
			if (a < an) {
				fout << v << " " << NG[v][i] << " (" << a << ") ";
				fout << "[" << NC[a] << "/" << NF[a] << "]";
				fout << endl;
			}
		}
	fout << "FLOW from " << sv << " to " << tv << ": "<< Fmax << endl;
}
extern vector<vertex> Nuv; 
extern vector<arc> Nue;
 void
    NF_tex( digraph& NG,  points& Np, arcs& NA, capacities &NC, flows &NF,   vertex sv, vertex tv, ofstream& ftex  )
{
    vertex vn = NG.size();
	arc an = 0;
	for (vertex v = 0; v < vn; v++)
		an = an + NG[v].size();
	an /= 2;
    ftex << endl;
	ftex << "\\begin{tikzpicture}" << endl;
    ftex << "[scale=3,every node/.style={fill=yellow!20}]" << endl;
    for( int v=0; v<vn; v++ ) {
        ftex << "\\node ("<<v<<") at (" << Np[v].first <<","<<-Np[v].second<<") {"<<v<<"};"<<endl;
    }
    ftex << "\\foreach \\from/\\to/\\arc in {";
    bool FIRST = true;
    for( int v=0; v<vn; v++ )
        for( unsigned int i=0; i<NG[v].size(); i++ ) {
            int u=NG[v][i];
			arc a = NA[vip(v, i)];
            if( a<an ) {
                if( FIRST ) 
                    FIRST = false;
                else 
                    ftex<<",";
                
				ftex << v <<"/" << u<<"/" << NF[a]<<"[" << NC[a] << "]";
            }
        }
        ftex << "}" << endl;
		
   ftex << "\\draw[->] (\\from) -- (\\to)" << endl;
   ftex << "node[sloped, midway, font=\\fontsize{10}{10}\\selectfont\\it]{\\arc};"<<endl;
   if( Nuv.size() ) {
	   ftex << "\\foreach \\from/\\to/\\arc in {";
	   FIRST = true;
	   vertex v = tv;
	   while (v != sv) {
		   arc a = Nue[v];
		   if (FIRST)
			   FIRST = false;
		   else
			   ftex << ",";
		   vertex u = Nuv[v];
		   if( a < an )
				ftex << u << "/" << v << "/" << NF[a] << "[" << NC[a] << "]";
		   else 
			   ftex << v << "/" << u << "/" << NF[a-an] << "[" << NC[a-an] << "]";
		   v = Nuv[v];
	   
	   }
	   ftex << "}" << endl;

	   ftex << "\\draw[red, ->] (\\from) -- (\\to)" << endl;
	   ftex << "node[sloped, midway, font=\\fontsize{10}{10}\\selectfont\\it]{\\arc};" << endl;
   }

   ftex << "\\end{tikzpicture}" << endl;
   ftex << "\\hspace{1cm}" << endl;
}


 // 
 //  Write digraph to fout
 //
 void
	 digraph_write(digraph& G, points &Gp, arcs &GA, ofstream& fout)
 {
	 //  Vertices number 
	 vertex vn = G.size();
	 //  Count vips number from adjacencies lists
	 arc an = 0;
	 for (vertex v = 0; v < vn; v++)
		 an = an + G[v].size();
	 an /= 2;
	 //  Write vips         
	 fout << "Network with " << vn << " vertices and " << an << " arcs " << endl;
	 fout << "Adjacencies lists" << endl;
	 for (vertex v = 0; v<vn; v++) {
		 fout << v << "\t:\t";
		 for (index i = 0; i<G[v].size(); i++) {
			 arc a = GA[vip(v, i)];
			 fout << "\t" << G[v][i] << " (" << a << ")";
			 if (DIGRAPH)
				 if (a < an) {
					 fout << "[out]";
				 }
				 else
					 fout << "[in]";
		 }
		 fout << endl;
	 }
	 if (DIGRAPH)
		 fout << "Arcs" << endl;
	 else
		 fout << "Edges" << endl;
	 for (vertex v = 0; v < vn; v++)
		 for (index i = 0; i < G[v].size(); i++) {
			 arc a = GA[vip(v, i)];
			 if (a < an) {
				 fout << v << " " << G[v][i] << " (" << a << ") ";
				 fout << endl;
			 }
		 }
 }
 void
	 digraph_tex(digraph& G, points &Gp, arcs &GA, ofstream& ftex)
 {
	 vertex vn = G.size();
	 arc an = 0;
	 for (vertex v = 0; v < vn; v++)
		 an = an + G[v].size();
	 an /= 2;
	 ftex << endl;
	 ftex << "\\begin{tikzpicture}" << endl;
	 ftex << "[scale=3,every node/.style={fill=yellow!20}]" << endl;
	 for (int v = 0; v<vn; v++) {
		 ftex << "\\node (" << v << ") at (" << Gp[v].first << "," << -Gp[v].second << ") {" << v << "};" << endl;
	 }
	 ftex << "\\foreach \\from/\\to/\\arc in {";
	 bool FIRST = true;
	 for (int v = 0; v<vn; v++)
		 for (unsigned int i = 0; i<G[v].size(); i++) {
			 int u = G[v][i];
			 arc a = GA[vip(v, i)];
			 if (a<an) {
				 if (FIRST)
					 FIRST = false;
				 else
					 ftex << ",";

				 ftex << v << "/" << u << "/" << a;
			 }
		 }
	 ftex << "}" << endl;

	 ftex << "\\draw (\\from) -- (\\to)" << endl;
	 ftex << "node[sloped, midway, font=\\fontsize{10}{10}\\selectfont\\it]{\\arc};" << endl;
	 ftex << "\\end{tikzpicture}" << endl;
	 ftex << "\\hspace{1cm}" << endl;
 }


 void
	 network_Init(digraph &NG, points &Np, arcs &NA, capacities &NC, flows &NF, vertex &sv, vertex &tv,
		 ofstream &ftex, ofstream &fout)
 {
	 string NFdat = NFstr;
	 NFdat += ".dat";
	 NF_read(NG, Np, NA, NC, NF, sv, tv, NFdat);
	 string NFout = NFstr;
	 NFout += ".out";
	 string NFtex = NFstr;
	 NFtex += ".tex";
	 fout.open(NFout);
	 ftex.open(NFtex);
	 NF_tex(NG, Np, NA, NC, NF, sv, tv, ftex);
	 NF_write(NG, Np, NA, NC, NF, sv, tv, fout);
 }

 void
	 bpnetwork_Init(digraph& NG, points& Np, arcs& NA, capacities& NC, flows& NF, vertex& sv, vertex& tv,
		 ofstream& ftex, ofstream& fout)
 {
	 string NFdat = BPNFstr;
	 NFdat += ".dat";
	 NF_read(NG, Np, NA, NC, NF, sv, tv, NFdat);
	 string NFout = BPNFstr;
	 NFout += ".out";
	 string NFtex = BPNFstr;
	 NFtex += ".tex";
	 fout.open(NFout);
	 ftex.open(NFtex);
	 NF_tex(NG, Np, NA, NC, NF, sv, tv, ftex);
	 NF_write(NG, Np, NA, NC, NF, sv, tv, fout);
 }
